# cython: auto_cpdef=True

"""Python code for writing AVRO files"""

# This code is a modified version of the code at
# http://svn.apache.org/viewvc/avro/trunk/lang/py/src/avro/ which is under
# Apache 2.0 license (http://www.apache.org/licenses/LICENSE-2.0)

from abc import ABC, abstractmethod
import json
from io import BytesIO
from os import urandom, SEEK_SET
import bz2
import lzma
import zlib
from typing import Union, IO, Iterable, Any, Optional, Dict

from .const import NAMED_TYPES
from .io.binary_encoder import BinaryEncoder
from .io.json_encoder import AvroJSONEncoder
from .validation import _validate
from .read import HEADER_SCHEMA, SYNC_SIZE, MAGIC, reader
from .logical_writers import LOGICAL_WRITERS
from .schema import extract_record_type, extract_logical_type, parse_schema
from ._write_common import _is_appendable
from .types import Schema, NamedSchemas


def write_null(encoder, datum, schema, named_schemas, fname, options):
    """null is written as zero bytes"""
    encoder.write_null()


def write_boolean(encoder, datum, schema, named_schemas, fname, options):
    """A boolean is written as a single byte whose value is either 0 (false) or
    1 (true)."""
    encoder.write_boolean(datum)


def write_int(encoder, datum, schema, named_schemas, fname, options):
    """int and long values are written using variable-length, zig-zag coding."""
    encoder.write_int(datum)


def write_long(encoder, datum, schema, named_schemas, fname, options):
    """int and long values are written using variable-length, zig-zag coding."""
    encoder.write_long(datum)


def write_float(encoder, datum, schema, named_schemas, fname, options):
    """A float is written as 4 bytes.  The float is converted into a 32-bit
    integer using a method equivalent to Java's floatToIntBits and then encoded
    in little-endian format."""
    encoder.write_float(datum)


def write_double(encoder, datum, schema, named_schemas, fname, options):
    """A double is written as 8 bytes.  The double is converted into a 64-bit
    integer using a method equivalent to Java's doubleToLongBits and then
    encoded in little-endian format."""
    encoder.write_double(datum)


def write_bytes(encoder, datum, schema, named_schemas, fname, options):
    """Bytes are encoded as a long followed by that many bytes of data."""
    encoder.write_bytes(datum)


def write_utf8(encoder, datum, schema, named_schemas, fname, options):
    """A string is encoded as a long followed by that many bytes of UTF-8
    encoded character data."""
    encoder.write_utf8(datum)


def write_crc32(encoder, datum):
    """A 4-byte, big-endian CRC32 checksum"""
    encoder.write_crc32(datum)


def write_fixed(encoder, datum, schema, named_schemas, fname, options):
    """Fixed instances are encoded using the number of bytes declared in the
    schema."""
    if len(datum) != schema["size"]:
        raise ValueError(
            f"data of length {len(datum)} does not match schema size: {schema}"
        )
    encoder.write_fixed(datum)


def write_enum(encoder, datum, schema, named_schemas, fname, options):
    """An enum is encoded by a int, representing the zero-based position of
    the symbol in the schema."""
    index = schema["symbols"].index(datum)
    encoder.write_enum(index)


def write_array(encoder, datum, schema, named_schemas, fname, options):
    """Arrays are encoded as a series of blocks.

    Each block consists of a long count value, followed by that many array
    items.  A block with count zero indicates the end of the array.  Each item
    is encoded per the array's item schema.

    If a block's count is negative, then the count is followed immediately by a
    long block size, indicating the number of bytes in the block.  The actual
    count in this case is the absolute value of the count written."""
    encoder.write_array_start()
    if len(datum) > 0:
        encoder.write_item_count(len(datum))
        dtype = schema["items"]
        for item in datum:
            write_data(encoder, item, dtype, named_schemas, fname, options)
            encoder.end_item()
    encoder.write_array_end()


def write_map(encoder, datum, schema, named_schemas, fname, options):
    """Maps are encoded as a series of blocks.

    Each block consists of a long count value, followed by that many key/value
    pairs.  A block with count zero indicates the end of the map.  Each item is
    encoded per the map's value schema.

    If a block's count is negative, then the count is followed immediately by a
    long block size, indicating the number of bytes in the block. The actual
    count in this case is the absolute value of the count written."""
    encoder.write_map_start()
    if len(datum) > 0:
        encoder.write_item_count(len(datum))
        vtype = schema["values"]
        for key, val in datum.items():
            encoder.write_utf8(key)
            write_data(encoder, val, vtype, named_schemas, fname, options)
    encoder.write_map_end()


def write_union(encoder, datum, schema, named_schemas, fname, options):
    """A union is encoded by first writing a long value indicating the
    zero-based position within the union of the schema of its value. The value
    is then encoded per the indicated schema within the union."""

    best_match_index = -1
    if isinstance(datum, tuple) and not options.get("disable_tuple_notation"):
        (name, datum) = datum
        for index, candidate in enumerate(schema):
            extracted_type = extract_record_type(candidate)
            if extracted_type in NAMED_TYPES:
                schema_name = candidate["name"]
            else:
                schema_name = extracted_type
            if name == schema_name:
                best_match_index = index
                break

        if best_match_index == -1:
            field = f"on field {fname}" if fname else ""
            msg = (
                f"provided union type name {name} not found in schema "
                + f"{schema} {field}"
            )
            raise ValueError(msg)
        index = best_match_index
    else:
        pytype = type(datum)
        most_fields = -1

        # All of Python's floating point values are doubles, so to
        # avoid loss of precision, we should always prefer 'double'
        # if we are forced to choose between float and double.
        #
        # If 'double' comes before 'float' in the union, then we'll immediately
        # choose it, and don't need to worry. But if 'float' comes before
        # 'double', we don't want to pick it.
        #
        # So, if we ever see 'float', we skim through the rest of the options,
        # just to see if 'double' is a possibility, because we'd prefer it.
        could_be_float = False

        for index, candidate in enumerate(schema):
            if could_be_float:
                if extract_record_type(candidate) == "double":
                    best_match_index = index
                    break
                else:
                    # Nothing except "double" is even worth considering.
                    continue

            if _validate(
                datum,
                candidate,
                named_schemas,
                raise_errors=False,
                field="",
                options=options,
            ):
                record_type = extract_record_type(candidate)
                if record_type == "record":
                    logical_type = extract_logical_type(candidate)
                    if logical_type:
                        prepare = LOGICAL_WRITERS.get(logical_type)
                        if prepare:
                            datum = prepare(datum, candidate)

                    candidate_fields = set(f["name"] for f in candidate["fields"])
                    datum_fields = set(datum)
                    fields = len(candidate_fields.intersection(datum_fields))
                    if fields > most_fields:
                        best_match_index = index
                        most_fields = fields
                elif record_type == "float":
                    best_match_index = index
                    # Continue in the loop, because it's possible that there's
                    # another candidate which has record type 'double'
                    could_be_float = True
                else:
                    best_match_index = index
                    break
        if best_match_index == -1:
            field = f"on field {fname}" if fname else ""
            raise ValueError(
                f"{repr(datum)} (type {pytype}) do not match {schema} {field}"
            )
        index = best_match_index

    # write data
    # TODO: There should be a way to give just the index
    encoder.write_index(index, schema[index])
    write_data(encoder, datum, schema[index], named_schemas, fname, options)


def write_record(encoder, datum, schema, named_schemas, fname, options):
    """A record is encoded by encoding the values of its fields in the order
    that they are declared. In other words, a record is encoded as just the
    concatenation of the encodings of its fields.  Field values are encoded per
    their schema."""
    extras = set(datum) - set(field["name"] for field in schema["fields"])
    if (options.get("strict") or options.get("strict_allow_default")) and extras:
        raise ValueError(
            f'record contains more fields than the schema specifies: {", ".join(extras)}'
        )
    for field in schema["fields"]:
        name = field["name"]
        field_type = field["type"]
        if name not in datum:
            if options.get("strict") or (
                options.get("strict_allow_default") and "default" not in field
            ):
                raise ValueError(
                    f"Field {name} is specified in the schema but missing from the record"
                )
            elif "default" not in field and "null" not in field_type:
                raise ValueError(f"no value and no default for {name}")
        datum_value = datum.get(name, field.get("default"))
        if field_type == "float" or field_type == "double":
            # Handle float values like "NaN"
            datum_value = float(datum_value)
        write_data(
            encoder,
            datum_value,
            field_type,
            named_schemas,
            name,
            options,
        )


WRITERS = {
    "null": write_null,
    "boolean": write_boolean,
    "string": write_utf8,
    "int": write_int,
    "long": write_long,
    "float": write_float,
    "double": write_double,
    "bytes": write_bytes,
    "fixed": write_fixed,
    "enum": write_enum,
    "array": write_array,
    "map": write_map,
    "union": write_union,
    "error_union": write_union,
    "record": write_record,
    "error": write_record,
}


def write_data(encoder, datum, schema, named_schemas, fname, options):
    """Write a datum of data to output stream.

    Parameters
    ----------
    encoder: encoder
        Type of encoder (e.g. binary or json)
    datum: object
        Data to write
    schema: dict
        Schema to use
    named_schemas: dict
        Mapping of fullname to schema definition
    """

    record_type = extract_record_type(schema)
    logical_type = extract_logical_type(schema)

    fn = WRITERS.get(record_type)
    if fn:
        if logical_type:
            prepare = LOGICAL_WRITERS.get(logical_type)
            if prepare:
                datum = prepare(datum, schema)
        try:
            return fn(encoder, datum, schema, named_schemas, fname, options)
        except TypeError as ex:
            if fname:
                raise TypeError(f"{ex} on field {fname}")
            raise
    else:
        return write_data(
            encoder, datum, named_schemas[record_type], named_schemas, "", options
        )


def write_header(encoder, metadata, sync_marker):
    header = {
        "magic": MAGIC,
        "meta": {key: value.encode() for key, value in metadata.items()},
        "sync": sync_marker,
    }
    write_data(encoder, header, HEADER_SCHEMA, {}, "", {})


def null_write_block(encoder, block_bytes, compression_level):
    """Write block in "null" codec."""
    encoder.write_long(len(block_bytes))
    encoder._fo.write(block_bytes)


def deflate_write_block(encoder, block_bytes, compression_level):
    """Write block in "deflate" codec."""
    # The first two characters and last character are zlib
    # wrappers around deflate data.
    if compression_level is not None:
        data = zlib.compress(block_bytes, compression_level)[2:-1]
    else:
        data = zlib.compress(block_bytes)[2:-1]
    encoder.write_long(len(data))
    encoder._fo.write(data)


def bzip2_write_block(encoder, block_bytes, compression_level):
    """Write block in "bzip2" codec."""
    data = bz2.compress(block_bytes)
    encoder.write_long(len(data))
    encoder._fo.write(data)


def xz_write_block(encoder, block_bytes, compression_level):
    """Write block in "xz" codec."""
    data = lzma.compress(block_bytes)
    encoder.write_long(len(data))
    encoder._fo.write(data)


BLOCK_WRITERS = {
    "null": null_write_block,
    "deflate": deflate_write_block,
    "bzip2": bzip2_write_block,
    "xz": xz_write_block,
}


def _missing_codec_lib(codec, library):
    def missing(encoder, block_bytes, compression_level):
        raise ValueError(
            f"{codec} codec is supported but you need to install {library}"
        )

    return missing


def snappy_write_block(encoder, block_bytes, compression_level):
    """Write block in "snappy" codec."""
    data = snappy.compress(block_bytes)
    encoder.write_long(len(data) + 4)  # for CRC
    encoder._fo.write(data)
    encoder.write_crc32(block_bytes)


try:
    import snappy
except ImportError:
    BLOCK_WRITERS["snappy"] = _missing_codec_lib("snappy", "python-snappy")
else:
    BLOCK_WRITERS["snappy"] = snappy_write_block


def zstandard_write_block(encoder, block_bytes, compression_level):
    """Write block in "zstandard" codec."""
    if compression_level is not None:
        data = zstd.ZstdCompressor(level=compression_level).compress(block_bytes)
    else:
        data = zstd.ZstdCompressor().compress(block_bytes)
    encoder.write_long(len(data))
    encoder._fo.write(data)


try:
    import zstandard as zstd
except ImportError:
    BLOCK_WRITERS["zstandard"] = _missing_codec_lib("zstandard", "zstandard")
else:
    BLOCK_WRITERS["zstandard"] = zstandard_write_block


def lz4_write_block(encoder, block_bytes, compression_level):
    """Write block in "lz4" codec."""
    data = lz4.block.compress(block_bytes)
    encoder.write_long(len(data))
    encoder._fo.write(data)


try:
    import lz4.block
except ImportError:
    BLOCK_WRITERS["lz4"] = _missing_codec_lib("lz4", "lz4")
else:
    BLOCK_WRITERS["lz4"] = lz4_write_block


class GenericWriter(ABC):
    def __init__(self, schema, metadata=None, validator=None, options={}):
        self._named_schemas = {}
        self.validate_fn = _validate if validator else None
        self.metadata = metadata or {}
        self.options = options

        # A schema of None is allowed when appending and when doing so the
        # self.schema will be updated later
        if schema is not None:
            self.schema = parse_schema(schema, self._named_schemas)

        if isinstance(schema, dict):
            schema = {
                key: value
                for key, value in schema.items()
                if key not in ("__fastavro_parsed", "__named_schemas")
            }
        elif isinstance(schema, list):
            schemas = []
            for s in schema:
                if isinstance(s, dict):
                    schemas.append(
                        {
                            key: value
                            for key, value in s.items()
                            if key
                            not in (
                                "__fastavro_parsed",
                                "__named_schemas",
                            )
                        }
                    )
                else:
                    schemas.append(s)
            schema = schemas

        self.metadata["avro.schema"] = json.dumps(schema)

    @abstractmethod
    def write(self, record):
        pass

    @abstractmethod
    def flush(self):
        pass


class Writer(GenericWriter):
    def __init__(
        self,
        fo: Union[IO, BinaryEncoder],
        schema: Schema,
        codec: str = "null",
        sync_interval: int = 1000 * SYNC_SIZE,
        metadata: Optional[Dict[str, str]] = None,
        validator: bool = False,
        sync_marker: bytes = b"",
        compression_level: Optional[int] = None,
        options: Dict[str, bool] = {},
    ):
        super().__init__(schema, metadata, validator, options)

        self.metadata["avro.codec"] = codec
        if isinstance(fo, BinaryEncoder):
            self.encoder = fo
        else:
            self.encoder = BinaryEncoder(fo)
        self.io = BinaryEncoder(BytesIO())
        self.block_count = 0
        self.sync_interval = sync_interval
        self.compression_level = compression_level

        if _is_appendable(self.encoder._fo):
            # Seed to the beginning to read the header
            self.encoder._fo.seek(0)
            avro_reader = reader(self.encoder._fo)
            header = avro_reader._header

            self._named_schemas = {}
            self.schema = parse_schema(avro_reader.writer_schema, self._named_schemas)

            codec = avro_reader.metadata.get("avro.codec", "null")

            self.sync_marker = header["sync"]

            # Seek to the end of the file
            self.encoder._fo.seek(0, 2)

            self.block_writer = BLOCK_WRITERS[codec]
        else:
            self.sync_marker = sync_marker or urandom(SYNC_SIZE)

            try:
                self.block_writer = BLOCK_WRITERS[codec]
            except KeyError:
                raise ValueError(f"unrecognized codec: {codec}")

            write_header(self.encoder, self.metadata, self.sync_marker)

    def dump(self):
        self.encoder.write_long(self.block_count)
        self.block_writer(self.encoder, self.io._fo.getvalue(), self.compression_level)
        self.encoder._fo.write(self.sync_marker)
        self.io._fo.truncate(0)
        self.io._fo.seek(0, SEEK_SET)
        self.block_count = 0

    def write(self, record):
        if self.validate_fn:
            self.validate_fn(
                record, self.schema, self._named_schemas, "", True, self.options
            )
        write_data(self.io, record, self.schema, self._named_schemas, "", self.options)
        self.block_count += 1
        if self.io._fo.tell() >= self.sync_interval:
            self.dump()

    def write_block(self, block):
        # Clear existing block if there are any records pending
        if self.io._fo.tell() or self.block_count > 0:
            self.dump()
        self.encoder.write_long(block.num_records)
        self.block_writer(self.encoder, block.bytes_.getvalue(), self.compression_level)
        self.encoder._fo.write(self.sync_marker)

    def flush(self):
        if self.io._fo.tell() or self.block_count > 0:
            self.dump()
        self.encoder._fo.flush()


class JSONWriter(GenericWriter):
    def __init__(
        self,
        fo: AvroJSONEncoder,
        schema: Schema,
        codec: str = "null",
        sync_interval: int = 1000 * SYNC_SIZE,
        metadata: Optional[Dict[str, str]] = None,
        validator: bool = False,
        sync_marker: bytes = b"",
        codec_compression_level: Optional[int] = None,
        options: Dict[str, bool] = {},
    ):
        super().__init__(schema, metadata, validator, options)

        self.encoder = fo
        self.encoder.configure(self.schema, self._named_schemas)

    def write(self, record):
        if self.validate_fn:
            self.validate_fn(
                record, self.schema, self._named_schemas, "", True, self.options
            )
        write_data(
            self.encoder, record, self.schema, self._named_schemas, "", self.options
        )

    def flush(self):
        self.encoder.flush()


def writer(
    fo: Union[IO, AvroJSONEncoder],
    schema: Schema,
    records: Iterable[Any],
    codec: str = "null",
    sync_interval: int = 1000 * SYNC_SIZE,
    metadata: Optional[Dict[str, str]] = None,
    validator: bool = False,
    sync_marker: bytes = b"",
    codec_compression_level: Optional[int] = None,
    *,
    strict: bool = False,
    strict_allow_default: bool = False,
    disable_tuple_notation: bool = False,
):
    """Write records to fo (stream) according to schema

    Parameters
    ----------
    fo
        Output stream
    schema
        Writer schema
    records
        Records to write. This is commonly a list of the dictionary
        representation of the records, but it can be any iterable
    codec
        Compression codec, can be 'null', 'deflate' or 'snappy' (if installed)
    sync_interval
        Size of sync interval
    metadata
        Header metadata
    validator
        If true, validation will be done on the records
    sync_marker
        A byte string used as the avro sync marker. If not provided, a random
        byte string will be used.
    codec_compression_level
        Compression level to use with the specified codec (if the codec
        supports it)
    strict
        If set to True, an error will be raised if records do not contain
        exactly the same fields that the schema states
    strict_allow_default
        If set to True, an error will be raised if records do not contain
        exactly the same fields that the schema states unless it is a missing
        field that has a default value in the schema
    disable_tuple_notation
        If set to True, tuples will not be treated as a special case. Therefore,
        using a tuple to indicate the type of a record will not work


    Example::

        from fastavro import writer, parse_schema

        schema = {
            'doc': 'A weather reading.',
            'name': 'Weather',
            'namespace': 'test',
            'type': 'record',
            'fields': [
                {'name': 'station', 'type': 'string'},
                {'name': 'time', 'type': 'long'},
                {'name': 'temp', 'type': 'int'},
            ],
        }
        parsed_schema = parse_schema(schema)

        records = [
            {u'station': u'011990-99999', u'temp': 0, u'time': 1433269388},
            {u'station': u'011990-99999', u'temp': 22, u'time': 1433270389},
            {u'station': u'011990-99999', u'temp': -11, u'time': 1433273379},
            {u'station': u'012650-99999', u'temp': 111, u'time': 1433275478},
        ]

        with open('weather.avro', 'wb') as out:
            writer(out, parsed_schema, records)

    The `fo` argument is a file-like object so another common example usage
    would use an `io.BytesIO` object like so::

        from io import BytesIO
        from fastavro import writer

        fo = BytesIO()
        writer(fo, schema, records)

    Given an existing avro file, it's possible to append to it by re-opening
    the file in `a+b` mode. If the file is only opened in `ab` mode, we aren't
    able to read some of the existing header information and an error will be
    raised. For example::

        # Write initial records
        with open('weather.avro', 'wb') as out:
            writer(out, parsed_schema, records)

        # Write some more records
        with open('weather.avro', 'a+b') as out:
            writer(out, None, more_records)

    Note: When appending, any schema provided will be ignored since the schema
    in the avro file will be re-used. Therefore it is convenient to just use
    None as the schema.
    """
    # Sanity check that records is not a single dictionary (as that is a common
    # mistake and the exception that gets raised is not helpful)
    if isinstance(records, dict):
        raise ValueError('"records" argument should be an iterable, not dict')

    output: Union[JSONWriter, Writer]
    if isinstance(fo, AvroJSONEncoder):
        output = JSONWriter(
            fo,
            schema,
            codec,
            sync_interval,
            metadata,
            validator,
            sync_marker,
            codec_compression_level,
            options={
                "strict": strict,
                "strict_allow_default": strict_allow_default,
                "disable_tuple_notation": disable_tuple_notation,
            },
        )
    else:
        output = Writer(
            BinaryEncoder(fo),
            schema,
            codec,
            sync_interval,
            metadata,
            validator,
            sync_marker,
            codec_compression_level,
            options={
                "strict": strict,
                "strict_allow_default": strict_allow_default,
                "disable_tuple_notation": disable_tuple_notation,
            },
        )

    for record in records:
        output.write(record)
    output.flush()


def schemaless_writer(
    fo: IO,
    schema: Schema,
    record: Any,
    *,
    strict: bool = False,
    strict_allow_default: bool = False,
    disable_tuple_notation: bool = False,
):
    """Write a single record without the schema or header information

    Parameters
    ----------
    fo
        Output file
    schema
        Schema
    record
        Record to write
    strict
        If set to True, an error will be raised if records do not contain
        exactly the same fields that the schema states
    strict_allow_default
        If set to True, an error will be raised if records do not contain
        exactly the same fields that the schema states unless it is a missing
        field that has a default value in the schema
    disable_tuple_notation
        If set to True, tuples will not be treated as a special case. Therefore,
        using a tuple to indicate the type of a record will not work


    Example::

        parsed_schema = fastavro.parse_schema(schema)
        with open('file', 'wb') as fp:
            fastavro.schemaless_writer(fp, parsed_schema, record)

    Note: The ``schemaless_writer`` can only write a single record.
    """
    named_schemas: NamedSchemas = {}
    schema = parse_schema(schema, named_schemas)

    encoder = BinaryEncoder(fo)
    write_data(
        encoder,
        record,
        schema,
        named_schemas,
        "",
        {
            "strict": strict,
            "strict_allow_default": strict_allow_default,
            "disable_tuple_notation": disable_tuple_notation,
        },
    )
    encoder.flush()



from typing import List, Optional, Set

from .repository import AbstractSchemaRepository
from .types import Schema, NamedSchemas

FINGERPRINT_ALGORITHMS: str

def load_schema(
    schema_path: str,
    *,
    repo: Optional[AbstractSchemaRepository] = ...,
    named_schemas: Optional[NamedSchemas] = ...,
    _write_hint: bool = ...,
    _injected_schemas: Optional[Set[str]] = ...,
) -> Schema: ...
def parse_schema(
    schema: Schema,
    named_schemas: Optional[NamedSchemas] = ...,
    *,
    expand: bool = ...,
    _write_hint: bool = ...,
    _force: bool = ...,
    _ignore_default_error: bool = ...,
) -> Schema: ...
def fullname(schema: Schema) -> str: ...
def expand_schema(schema: Schema) -> Schema: ...
def load_schema_ordered(
    ordered_schemas: List[str], *, _write_hint: bool = ...
) -> Schema: ...
def to_parsing_canonical_form(schema: Schema) -> str: ...
def fingerprint(parsing_canonical_form: str, algorithm: str) -> str: ...



def _is_appendable(file_like):
    if file_like.seekable() and file_like.tell() != 0:
        if "<stdout>" == getattr(file_like, "name", ""):
            # In OSX, sys.stdout is seekable and has a non-zero tell() but
            # we wouldn't want to append to a stdout. In the python REPL,
            # sys.stdout is named `<stdout>`
            return False
        if file_like.readable():
            return True
        else:
            raise ValueError(
                "When appending to an avro file you must use the "
                + "'a+' mode, not just 'a'"
            )
    else:
        return False



# -*- coding: utf-8 -*-
from fastavro.validation import (
    ValidationError,
    ValidationErrorData,
    validate,
    validate_many,
)
from fastavro import parse_schema
import pytest
import numpy as np
from datetime import datetime

schema = {
    "fields": [
        {"name": "str_null", "type": ["null", "string"]},
        {"name": "str", "type": "string"},
        {"name": "integ_null", "type": ["null", "int"]},
        {"name": "integ", "type": "int"},
    ],
    "namespace": "namespace",
    "name": "missingerror",
    "type": "record",
}


# TODO: Add more test for all types and combinations


def validation_boolean(schema, *records):
    return validate_many(records, schema, raise_errors=False)


def validation_raise(schema, *records):
    return validate_many(records, schema, raise_errors=True)


def test_validate_string_in_int_raises():
    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": "str",
            "integ": 21,
        }
    ]

    with pytest.raises(ValidationError) as exc:
        validation_raise(schema, *records)

    for error in exc.value.errors:
        expected_type = error.schema
        assert expected_type in ["null", "int"]
        assert error.field == "namespace.missingerror.integ_null"


def test_validate_string_in_int_false():
    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": "str",
            "integ": 21,
        }
    ]

    assert validation_boolean(schema, *records) is False


def test_validate_true():
    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": 21,
            "integ": 21,
        },
        {
            "str_null": None,
            "str": "str",
            "integ_null": None,
            "integ": 21,
        },
    ]

    assert validation_boolean(schema, *records) is True
    validation_raise(schema, *records)


def test_validate_string_in_int_null_raises():
    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": 11,
            "integ": "str",
        }
    ]

    with pytest.raises(ValidationError) as exc:
        validation_raise(schema, *records)

    for error in exc.value.errors:
        expected_type = error.schema
        assert expected_type == "int"
        assert error.field == "namespace.missingerror.integ"


def test_validate_string_in_int_null_false():
    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": 11,
            "integ": "str",
        }
    ]

    assert validation_boolean(schema, *records) is False


def test_validate_int_in_string_null_raises():
    records = [
        {
            "str_null": 11,
            "str": "str",
            "integ_null": 21,
            "integ": 21,
        }
    ]

    with pytest.raises(ValidationError) as exc:
        validation_raise(schema, *records)

    for error in exc.value.errors:
        expected_type = error.schema
        assert expected_type in ["string", "null"]
        assert error.field == "namespace.missingerror.str_null"


def test_validate_int_in_string_null_false():
    records = [
        {
            "str_null": 11,
            "str": "str",
            "integ_null": 21,
            "integ": 21,
        }
    ]
    assert validation_boolean(schema, *records) is False


def test_validate_int_in_string_raises():
    records = [
        {
            "str_null": "str",
            "str": 11,
            "integ_null": 21,
            "integ": 21,
        }
    ]

    with pytest.raises(ValidationError) as exc:
        validation_raise(schema, *records)

    for error in exc.value.errors:
        expected_type = error.schema
        assert expected_type == "string"
        assert error.field == "namespace.missingerror.str"


def test_validate_int_in_string_false():
    records = [
        {
            "str_null": "str",
            "str": 11,
            "integ_null": 21,
            "integ": 21,
        }
    ]

    assert validation_boolean(schema, *records) is False


def test_validate_null_in_string_raises():
    records = [
        {
            "str_null": "str",
            "str": None,
            "integ_null": 21,
            "integ": 21,
        }
    ]

    with pytest.raises((ValidationError,)):
        validation_raise(schema, *records)


def test_validate_null_in_string_false():
    records = [
        {
            "str_null": "str",
            "str": None,
            "integ_null": 21,
            "integ": 21,
        }
    ]

    assert validation_boolean(schema, *records) is False


def test_validate_unicode_in_string_does_not_raise():
    """https://github.com/fastavro/fastavro/issues/269"""
    non_ascii = "日本語"

    records = [
        {
            "str_null": non_ascii,
            "str": "str",
            "integ_null": 21,
            "integ": 21,
        }
    ]

    validation_raise(schema, *records)

    records = [
        {
            "str_null": "str",
            "str": "str",
            "integ_null": 21,
            "integ": non_ascii,
        }
    ]

    with pytest.raises(ValidationError) as exc:
        validation_raise(schema, *records)

    for error in exc.value.errors:
        assert error.datum == non_ascii


def test_validate_error_raises():
    with pytest.raises(ValidationError):
        raise ValidationError()

    error = ValidationErrorData(10, "string", "test1")
    msg = "test1 is <10> of type <class 'int'> expected string"
    assert msg in str(error)


def test_validate_error_none_field():
    error = ValidationErrorData(10, "string", None)
    msg = " is <10> of type <class 'int'> expected string"
    assert msg in str(error)


def test_validator_numeric():
    for datum, schema in [
        (1, "int"),
        (1, "long"),
        (1.0, "float"),
        (1.0, "double"),
        (1, "float"),
        (1, "double"),
    ]:
        validate(datum, schema)

    for datum, schema in [
        (1.0, "int"),
        (1.0, "long"),
        ("1.0", "float"),
        ("1.0", "double"),
        ("1", "float"),
        ("1", "double"),
        (True, "int"),
        (True, "long"),
        (True, "float"),
        (True, "double"),
        (False, "int"),
        (False, "long"),
        (False, "float"),
        (False, "double"),
        (datetime(2020, 1, 1), "int"),
        (datetime(2020, 1, 1), "long"),
    ]:
        with pytest.raises(ValidationError):
            validate(datum, schema)
            pytest.fail(f"{datum} should not validate as {schema}")


def test_validator_logical():
    """https://github.com/fastavro/fastavro/issues/365"""
    for datum, schema in [
        (1, {"type": "long", "logicalType": "timestamp-micros"}),
    ]:
        validate(datum, schema)

    for datum, schema in [
        ("foo", {"type": "long", "logicalType": "timestamp-micros"}),
    ]:
        with pytest.raises(ValidationError):
            validate(datum, schema)
            pytest.fail(f"{datum} should not validate as {schema}")


def test_validate_array():
    my_schema = {
        "fields": [
            {
                "name": "array",
                "type": {
                    "type": "array",
                    "items": "string",
                },
            },
        ],
        "namespace": "namespace",
        "name": "test_validate_array",
        "type": "record",
    }

    datum = {"array": [1]}
    with pytest.raises(ValidationError) as exc:
        validate(datum, my_schema)

    for error in exc.value.errors:
        assert error.field == "namespace.test_validate_array.array"


def test_validate_map():
    my_schema = {
        "fields": [
            {
                "name": "map",
                "type": {
                    "type": "map",
                    "values": "string",
                },
            },
        ],
        "namespace": "namespace",
        "name": "test_validate_map",
        "type": "record",
    }

    datum = {"map": {"key": 1}}
    with pytest.raises(ValidationError) as exc:
        validate(datum, my_schema)

    for error in exc.value.errors:
        assert error.field == "namespace.test_validate_map.map"


def test_validator_numeric_numpy():
    np_ints = [
        np.int_,
        np.intc,
        np.intp,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
    ]

    np_floats = [
        np.float_,
        np.float16,
        np.float32,
        np.float64,
    ]

    schema_ints = ["int", "long"]

    schema_floats = ["float", "double"]

    # all these should work
    for nptype, schema in zip(np_ints, schema_ints):
        validate(nptype(1), schema)

    for nptype, schema in zip(np_ints, schema_floats):
        validate(nptype(1), schema)

    for nptype, schema in zip(np_floats, schema_floats):
        validate(nptype(1), schema)

    # these shouldn't work
    for nptype, schema in zip(np_floats, schema_ints):
        with pytest.raises(ValidationError):
            validate(nptype(1), schema)


def test_validate_with_unparsed_schema():
    """https://github.com/fastavro/fastavro/issues/416"""
    schema = {
        "type": "record",
        "name": "my_schema_v1",
        "fields": [
            {
                "name": "field1",
                "type": {
                    "type": "record",
                    "name": "my_field_type",
                    "fields": [{"name": "sub_field", "type": "string"}],
                },
            },
            {"name": "field2", "type": "my_field_type"},
        ],
    }

    datum = {"field1": {"sub_field": "foo"}, "field2": {"sub_field": "bar"}}

    validate(datum, schema)


def test_validate_clobbering_SCHEMA_DEFS():
    """https://github.com/fastavro/fastavro/issues/416"""
    schema_1 = {
        "type": "record",
        "name": "my_schema_v1",
        "fields": [
            {
                "name": "field1",
                "type": {
                    "type": "record",
                    "name": "my_field_type",
                    "fields": [{"name": "sub_field", "type": "string"}],
                },
            },
            {"name": "field2", "type": "my_field_type"},
        ],
    }
    parsed_schema_1 = parse_schema(schema_1)

    datum_1 = {"field1": {"sub_field": "foo"}, "field2": {"sub_field": "bar"}}

    validate(datum_1, parsed_schema_1)

    schema_2 = {
        "type": "record",
        "name": "my_schema_v2",
        "fields": [
            {
                "name": "field1",
                "type": {
                    "type": "record",
                    "name": "my_field_type",
                    "fields": [{"name": "sub_field", "type": "int"}],
                },
            }
        ],
    }
    parse_schema(schema_2)

    validate(datum_1, parsed_schema_1)


def test_enum_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_enum_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "enum",
                    "name": "my_enum",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {
                "name": "test2",
                "type": "my_enum",
            },
        ],
    }

    record = {"test1": "FOO", "test2": "BAR"}
    parsed_schema = parse_schema(schema)
    validate(record, parsed_schema)


def test_fixed_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_fixed_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "fixed",
                    "name": "my_fixed",
                    "size": 4,
                },
            },
            {
                "name": "test2",
                "type": "my_fixed",
            },
        ],
    }

    record = {"test1": b"1234", "test2": b"4321"}
    parsed_schema = parse_schema(schema)
    validate(record, parsed_schema)


def test_record_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_record_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "record",
                    "name": "my_record",
                    "fields": [
                        {
                            "name": "field1",
                            "type": "string",
                        }
                    ],
                },
            },
            {
                "name": "test2",
                "type": "my_record",
            },
        ],
    }

    record = {"test1": {"field1": "foo"}, "test2": {"field1": "bar"}}
    parsed_schema = parse_schema(schema)
    validate(record, parsed_schema)


def test_record_name_with_named_type_in_union():
    schema = {
        "type": "record",
        "name": "my_record",
        "fields": [
            {
                "name": "my_1st_union",
                "type": [
                    {
                        "name": "foo",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                    {
                        "name": "bar",
                        "type": "record",
                        "fields": [{"name": "some_field", "type": "int"}],
                    },
                ],
            },
            {"name": "my_2nd_union", "type": ["foo", "bar"]},
        ],
    }

    records = [
        {
            "my_1st_union": ("foo", {"some_field": 1}),
            "my_2nd_union": ("bar", {"some_field": 2}),
        }
    ]

    parsed_schema = parse_schema(schema)
    validate_many(records, parsed_schema)


def test_validate_should_not_parse_schema_if_it_was_parsed_already():
    named_schemas = {}
    parse_schema(
        {"name": "B", "type": "record", "fields": [{"name": "bar", "type": "string"}]},
        named_schemas,
    )

    a_schema = parse_schema(
        {"name": "A", "type": "record", "fields": [{"name": "b", "type": "B"}]},
        named_schemas,
    )

    records = [{"b": {"bar": "bar"}}]

    validate_many(records, a_schema)
    validate(records[0], a_schema)


def test_validate_strict():
    """https://github.com/fastavro/fastavro/issues/423"""
    schema = {
        "type": "record",
        "name": "test_validate_strict",
        "fields": [
            {"name": "normal_field", "type": "string"},
            {"name": "normal_union_field", "type": ["int", "float"]},
            {"name": "nullable_union_field", "type": ["null", "string"]},
        ],
    }

    record = {"normal_field": "a", "normal_union_field": 1}
    parsed_schema = parse_schema(schema)
    with pytest.raises(ValidationError):
        validate(record, parsed_schema, strict=True)

    with pytest.raises(ValidationError):
        validate_many([record], parsed_schema, strict=True)



from io import BytesIO
import fastavro

import pytest

schema = {
    "type": "record",
    "name": "test_block_iteration",
    "fields": [
        {"name": "nullable_str", "type": ["string", "null"]},
        {"name": "str_field", "type": "string"},
        {"name": "int_field", "type": "int"},
    ],
}


def make_records(num_records=2000):
    return [
        {
            "nullable_str": None if i % 3 == 0 else f"{i}-{i}",
            "str_field": f"{i} {i} {i}",
            "int_field": i * 10,
        }
        for i in range(num_records)
    ]


def make_blocks(num_records=2000, codec="null"):
    records = make_records(num_records)

    new_file = BytesIO()
    fastavro.writer(new_file, schema, records, codec=codec)

    new_file.seek(0)
    block_reader = fastavro.block_reader(new_file, schema)

    blocks = list(block_reader)

    new_file.close()

    return blocks, records


@pytest.mark.parametrize(
    "source_codec,output_codec",
    [
        ("null", "null"),
        ("deflate", "deflate"),
        ("null", "deflate"),
        ("deflate", "null"),
    ],
)
def test_check_concatenate(source_codec, output_codec):
    blocks1, records1 = make_blocks(codec=source_codec)
    blocks2, records2 = make_blocks(codec=source_codec)

    new_file = BytesIO()
    w = fastavro.write.Writer(new_file, schema, codec=output_codec)
    for block in blocks1:
        w.write_block(block)
    for block in blocks2:
        w.write_block(block)

    # Read the file back to make sure we get back the same stuff
    new_file.seek(0)
    new_records = list(fastavro.reader(new_file, schema))
    assert new_records == records1 + records2



from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    IO,
    Optional,
    Union,
    Tuple,
    Generic,
    TypeVar,
)
from .io.json_decoder import AvroJSONDecoder
from .types import AvroMessage, Schema

T = TypeVar("T")

class file_reader(Generic[T]):
    decoder: Union[IO, AvroJSONDecoder]
    return_record_name: bool
    metadata: Dict[str, str]
    codec: str
    reader_schema: Optional[Schema]
    writer_schema: Schema
    _header: Dict[str, Any]
    def __init__(
        self,
        fo_or_decoder: Union[IO, AvroJSONDecoder],
        reader_schema: Optional[Schema] = ...,
        return_record_name: bool = ...,
    ): ...
    def __iter__(self) -> Iterator[T]: ...
    def __next__(self) -> T: ...

class reader(file_reader[AvroMessage]):
    def __init__(
        self,
        fo: Union[IO, AvroJSONDecoder],
        reader_schema: Optional[Schema] = ...,
        return_record_name: bool = ...,
    ): ...

class block_reader(file_reader[Block]):
    def __init__(
        self,
        fo: IO,
        reader_schema: Optional[Schema] = ...,
        return_record_name: bool = ...,
    ): ...

class Block:
    num_records: int
    writer_schema: Dict
    reader_schema: Dict
    offset: int
    size: int
    def __init__(
        self,
        bytes_: bytes,
        num_records: int,
        codec: str,
        reader_schema: Dict,
        writer_schema: Dict,
        offset: int,
        size: int,
        return_record_name: bool,
    ): ...
    def __iter__(self) -> Iterator[AvroMessage]: ...
    def __str__(self) -> str: ...

def schemaless_reader(
    fo: IO,
    writer_schema: Schema,
    reader_schema: Optional[Schema],
    return_record_name: bool = ...,
) -> AvroMessage: ...
def is_avro(path_or_buffer: Union[str, IO]) -> bool: ...

BLOCK_READERS: Dict[str, Callable]



from os.path import join, abspath, dirname
import pytest
from fastavro.repository.base import SchemaRepositoryError
from fastavro.repository.flat_dict import FlatDictRepository


def test_load_returns_parsed_value():
    repo_dir = join(abspath(dirname(__file__)), "flat_dict_test")
    repo = FlatDictRepository(repo_dir)

    result = repo.load("Valid")
    expected = {
        "name": "Valid",
        "type": "record",
        "fields": [{"name": "foo", "type": "string"}],
    }
    assert expected == result


def test_load_raises_error_if_file_does_not_exist():
    repo_dir = join(abspath(dirname(__file__)), "flat_dict_test")
    repo = FlatDictRepository(repo_dir)

    with pytest.raises(SchemaRepositoryError) as err:
        repo.load("MissingFile")

    assert str(err.value) == "Failed to load 'MissingFile' schema"


def test_load_raises_error_if_file_is_not_valid_json():
    repo_dir = join(abspath(dirname(__file__)), "flat_dict_test")
    repo = FlatDictRepository(repo_dir)

    with pytest.raises(SchemaRepositoryError) as err:
        repo.load("InvalidJson")

    assert str(err.value) == "Failed to parse 'InvalidJson' schema"



from io import BytesIO
import fastavro
from tempfile import NamedTemporaryFile

schema = {
    "type": "record",
    "name": "test_block_iteration",
    "fields": [
        {"name": "nullable_str", "type": ["string", "null"]},
        {"name": "str_field", "type": "string"},
        {"name": "int_field", "type": "int"},
    ],
}


def make_records(num_records=2000):
    return [
        {
            "nullable_str": None if i % 3 == 0 else f"{i}-{i}",
            "str_field": f"{i} {i} {i}",
            "int_field": i * 10,
        }
        for i in range(num_records)
    ]


def make_blocks(num_records=2000, codec="null", write_to_disk=False):
    records = make_records(num_records)

    new_file = NamedTemporaryFile() if write_to_disk else BytesIO()
    fastavro.writer(new_file, schema, records, codec=codec)
    bytes = new_file.tell()

    new_file.seek(0)
    block_reader = fastavro.block_reader(new_file, schema)

    blocks = list(block_reader)

    new_file.close()

    return blocks, records, bytes


def check_block(block, num_bytes, num_records, records, codec, offset, size):
    block_records = list(block)

    assert len(block_records) == num_records
    assert block.codec == codec
    assert block.reader_schema == schema
    assert block.writer_schema == schema
    assert block.offset == offset
    assert block.size == size
    assert block_records == records


def check_round_trip(write_to_disk):
    blocks, records, bytes = make_blocks(write_to_disk=write_to_disk)

    assert bytes == 46007

    check_block(blocks[0], 16004, 811, records[:811], "null", 247, 16025)
    check_block(blocks[1], 16016, 656, records[811 : 811 + 656], "null", 16272, 16037)
    check_block(blocks[2], 13677, 533, records[811 + 656 :], "null", 32309, 13698)


def check_round_trip_deflated(write_to_disk):
    blocks, records, bytes = make_blocks(codec="deflate", write_to_disk=write_to_disk)

    assert bytes == 16543

    check_block(blocks[0], 16004, 811, records[:811], "deflate", 250, 6242)
    check_block(blocks[1], 16016, 656, records[811 : 811 + 656], "deflate", 6492, 5624)
    check_block(blocks[2], 13677, 533, records[811 + 656 :], "deflate", 12116, 4427)


def test_block_iteration_disk():
    check_round_trip(write_to_disk=True)


def test_block_iteration_memory():
    check_round_trip(write_to_disk=False)


def test_block_iteration_deflated_disk():
    check_round_trip_deflated(write_to_disk=True)


def test_block_iteration_deflated_memory():
    check_round_trip_deflated(write_to_disk=False)



# Some basic unit and integration tests for the 'fastavro' CLI
#
# std imports
import os
import sys
import json
import subprocess

data_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "avro-files")

main_py = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), os.pardir, "fastavro", "__main__.py"
)


def test_cli_record_output():
    # given,
    given_avro_input = os.path.join(data_dir, "weather.avro")
    given_cmd_args = [sys.executable, main_py, given_avro_input]
    expected_data = [
        {"station": "011990-99999", "time": -619524000000, "temp": 0},
        {"station": "011990-99999", "time": -619506000000, "temp": 22},
        {"station": "011990-99999", "time": -619484400000, "temp": -11},
        {"station": "012650-99999", "time": -655531200000, "temp": 111},
        {"station": "012650-99999", "time": -655509600000, "temp": 78},
    ]

    # exercise,
    result_output = subprocess.check_output(given_cmd_args).decode().splitlines()
    data = [json.loads(result_line_out) for result_line_out in result_output]

    # verify
    assert data == expected_data


def test_cli_stream_input():
    # given,
    given_avro_input = os.path.join(data_dir, "weather.avro")
    given_stdin_stream = open(given_avro_input, "rb")
    given_cmd_args = [sys.executable, main_py, "-"]
    expected_data = [
        {"station": "011990-99999", "time": -619524000000, "temp": 0},
        {"station": "011990-99999", "time": -619506000000, "temp": 22},
        {"station": "011990-99999", "time": -619484400000, "temp": -11},
        {"station": "012650-99999", "time": -655531200000, "temp": 111},
        {"station": "012650-99999", "time": -655509600000, "temp": 78},
    ]

    # exercise,
    result_output = (
        subprocess.check_output(given_cmd_args, stdin=given_stdin_stream)
        .decode()
        .splitlines()
    )
    data = [json.loads(result_line_out) for result_line_out in result_output]

    # verify
    assert data == expected_data


def test_cli_arg_metadata():
    # given,
    given_avro_input = os.path.join(data_dir, "testDataFileMeta.avro")
    given_cmd_args = [sys.executable, main_py, "--metadata", given_avro_input]
    expected_metadata = {"hello": "bar"}

    # exercise,
    result_output = subprocess.check_output(given_cmd_args).decode()
    data = json.loads(result_output)

    # verify
    assert data == expected_metadata


def test_cli_arg_schema():
    # given,
    given_avro_input = os.path.join(data_dir, "weather.avro")
    given_cmd_args = [sys.executable, main_py, "--schema", given_avro_input]
    expected_schema = {
        "type": "record",
        "name": "Weather",
        "namespace": "test",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
        ],
        "doc": "A weather reading.",
    }

    # exercise,
    result_output = subprocess.check_output(given_cmd_args).decode()
    data = json.loads(result_output)

    # verify
    assert data == expected_schema


def test_cli_arg_codecs():
    # given,
    given_cmd_args = [sys.executable, main_py, "--codecs"]
    default_codecs = ("deflate", "null")

    # exercise,
    result_output = subprocess.check_output(given_cmd_args).decode()
    result_codecs = [
        line.strip() for line in result_output.splitlines() if line.strip()
    ]

    for codec in default_codecs:
        assert codec in result_codecs



from copy import deepcopy
from io import StringIO
import json

import pytest

from fastavro import json_writer, json_reader
from fastavro.schema import parse_schema
from fastavro.validation import ValidationError
from fastavro.io.json_decoder import AvroJSONDecoder
from fastavro.io.json_encoder import AvroJSONEncoder
from fastavro.io.symbols import MapKeyMarker, String


def roundtrip(schema, records, *, writer_kwargs={}):
    new_file = StringIO()
    json_writer(new_file, schema, records, **writer_kwargs)
    new_file.seek(0)

    new_records = list(json_reader(new_file, schema))
    return new_records


def test_json():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "null", "type": "null"},
            {"name": "boolean", "type": "boolean"},
            {"name": "string", "type": "string"},
            {"name": "bytes", "type": "bytes"},
            {"name": "int", "type": "int"},
            {"name": "long", "type": "long"},
            {"name": "float", "type": "float"},
            {"name": "double", "type": "double"},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            },
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {"name": "array", "type": {"type": "array", "items": "string"}},
            {"name": "map", "type": {"type": "map", "values": "int"}},
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
            },
        ],
    }

    records = [
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": None,
            "enum": "BAR",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {"sub_int": 123},
        },
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": 321,
            "enum": "BAR",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {"sub_int": 123},
        },
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": {"union_record_field": "union_field"},
            "enum": "BAR",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {"sub_int": 123},
        },
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_more_than_one_record():
    schema = {
        "type": "record",
        "name": "test_more_than_one_record",
        "namespace": "test",
        "fields": [
            {"name": "string", "type": "string"},
            {"name": "int", "type": "int"},
        ],
    }

    records = [{"string": "foo", "int": 1}, {"string": "bar", "int": 2}]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_encoded_union_output():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            }
        ],
    }

    # A null value is encoded as just null
    records = [{"union": None}]
    new_file = StringIO()
    json_writer(new_file, schema, records)
    assert new_file.getvalue().strip() == json.dumps({"union": None})

    # A non-null, non-named type is encoded as an object with a key for the
    # type
    records = [{"union": 321}]
    new_file = StringIO()
    json_writer(new_file, schema, records)
    assert new_file.getvalue().strip() == json.dumps({"union": {"int": 321}})

    # A non-null, named type is encoded as an object with a key for the name
    records = [{"union": {"union_record_field": "union_field"}}]
    new_file = StringIO()
    json_writer(new_file, schema, records)
    expected = json.dumps(
        {"union": {"test.union_record": {"union_record_field": "union_field"}}}
    )
    assert new_file.getvalue().strip() == expected


def test_union_output_without_type():
    """https://github.com/fastavro/fastavro/issues/420"""
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            }
        ],
    }

    # A null value is encoded as just null
    records = [{"union": None}]
    new_file = StringIO()
    json_writer(new_file, schema, records, write_union_type=False)
    assert new_file.getvalue().strip() == json.dumps({"union": None})

    # A non-null, non-named type is encoded as just the value
    records = [{"union": 321}]
    new_file = StringIO()
    json_writer(new_file, schema, records, write_union_type=False)
    assert new_file.getvalue().strip() == json.dumps({"union": 321})

    # A non-null, named type is encoded as an object
    records = [{"union": {"union_record_field": "union_field"}}]
    new_file = StringIO()
    json_writer(new_file, schema, records, write_union_type=False)
    expected = json.dumps({"union": {"union_record_field": "union_field"}})
    assert new_file.getvalue().strip() == expected


def test_union_string_and_bytes():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "union", "type": ["string", "bytes"]}],
    }

    records = [{"union": "asdf"}, {"union": b"asdf"}]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_simple_type():
    schema = {"type": "string"}

    records = ["foo", "bar"]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_array_type_simple():
    schema = {"type": "array", "items": "string"}

    records = [["foo", "bar"], ["a", "b"]]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_array_type_records():
    schema = {
        "type": "array",
        "items": {
            "type": "record",
            "name": "test_array_type",
            "fields": [
                {"name": "field1", "type": "string"},
                {"name": "field2", "type": "int"},
            ],
        },
    }

    records = [[{"field1": "foo", "field2": 1}], [{"field1": "bar", "field2": 2}]]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_empty_maps():
    """https://github.com/fastavro/fastavro/issues/380"""
    schema = {"type": "map", "values": "int"}

    records = [{"foo": 1}, {}]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_empty_arrays():
    """https://github.com/fastavro/fastavro/issues/380"""
    schema = {"type": "array", "items": "int"}

    records = [[1], []]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_union_in_array():
    """https://github.com/fastavro/fastavro/issues/399"""
    schema = {
        "type": "array",
        "items": [
            {
                "type": "record",
                "name": "rec1",
                "fields": [{"name": "field1", "type": ["string", "null"]}],
            },
            {
                "type": "record",
                "name": "rec2",
                "fields": [{"name": "field2", "type": ["string", "null"]}],
            },
            "null",
        ],
    }

    records = [
        [{"field1": "foo"}, {"field2": None}, None],
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_union_in_array2():
    """https://github.com/fastavro/fastavro/issues/399"""
    schema = {
        "type": "record",
        "name": "Inbox",
        "fields": [
            {"type": "string", "name": "id"},
            {"type": "string", "name": "msg_title"},
            {
                "name": "msg_content",
                "type": {
                    "type": "array",
                    "items": [
                        {
                            "type": "record",
                            "name": "LimitedTime",
                            "fields": [
                                {
                                    "type": ["string", "null"],
                                    "name": "type",
                                    "default": "now",
                                }
                            ],
                        },
                        {
                            "type": "record",
                            "name": "Text",
                            "fields": [{"type": ["string", "null"], "name": "text"}],
                        },
                    ],
                },
            },
        ],
    }

    records = [
        {
            "id": 1234,
            "msg_title": "Hi",
            "msg_content": [{"type": "now"}, {"text": "hi from here!"}],
        },
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_union_in_map():
    """https://github.com/fastavro/fastavro/issues/399"""
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {
                "name": "map",
                "type": {"type": "map", "values": ["string", "null"]},
            }
        ],
    }

    records = [{"map": {"c": "1", "d": None}}]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_with_dependent_schema():
    """Tests a schema with dependent schema
    https://github.com/fastavro/fastavro/issues/418"""
    dependency = {
        "type": "record",
        "name": "Dependency",
        "namespace": "test",
        "fields": [{"name": "_name", "type": "string"}],
    }

    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "_name", "type": "string"},
            {"name": "_dependency", "type": "Dependency"},
        ],
    }

    records = [{"_name": "parent", "_dependency": {"_name": "child"}}]

    parsed_schema = parse_schema([dependency, schema])

    new_records = roundtrip(parsed_schema, records)
    assert records == new_records


def test_enum_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_enum_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {"type": "enum", "name": "my_enum", "symbols": ["FOO", "BAR"]},
            },
            {"name": "test2", "type": "my_enum"},
        ],
    }

    records = [{"test1": "FOO", "test2": "BAR"}]
    parsed_schema = parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_fixed_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_fixed_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {"type": "fixed", "name": "my_fixed", "size": 4},
            },
            {"name": "test2", "type": "my_fixed"},
        ],
    }

    records = [{"test1": b"1234", "test2": b"4321"}]
    parsed_schema = parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_record_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_record_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "record",
                    "name": "my_record",
                    "fields": [{"name": "field1", "type": "string"}],
                },
            },
            {"name": "test2", "type": "my_record"},
        ],
    }

    records = [{"test1": {"field1": "foo"}, "test2": {"field1": "bar"}}]
    parsed_schema = parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_default_union_values():
    """https://github.com/fastavro/fastavro/issues/485"""
    schema = {
        "type": "record",
        "name": "User",
        "fields": [
            {"name": "name", "type": "string"},
            {"name": "age", "type": "long"},
            {
                "name": "pets",
                "type": {"type": "array", "items": "string"},
            },
            {
                "name": "accounts",
                "type": {"type": "map", "values": "long"},
            },
            {
                "name": "favorite_colors",
                "type": {
                    "type": "enum",
                    "name": "favorite_color",
                    "symbols": ["BLUE", "YELLOW", "GREEN"],
                },
            },
            {"name": "country", "type": ["string", "null"], "default": "Argentina"},
            {"name": "address", "type": ["null", "string"], "default": None},
        ],
        "doc": "An User",
        "namespace": "User.v1",
        "aliases": ["user-v1", "super user"],
    }

    record = {
        "name": "MgXqfDAqzbgJSTTHDXtN",
        "age": 551,
        "pets": ["aRvwODwbOWfrkxYYkJiI"],
        "accounts": {"DQSZRzofFrNCiOhhIOvX": 4431},
        "favorite_colors": "GREEN",
        "address": {"string": "YgmVDKhXctMgODKkhNHJ"},
    }

    new_file = StringIO(json.dumps(record))
    read_record = next(json_reader(new_file, schema))

    assert read_record["country"] == "Argentina"


def test_all_default_values():
    """https://github.com/fastavro/fastavro/issues/485"""
    default_boolean = True
    default_string = "default_string"
    default_bytes = "default_bytes"
    default_int = -1
    default_long = -2
    default_float = 1.1
    default_double = 2.2
    default_fixed = "12345"
    default_union = None
    default_enum = "FOO"
    default_array = ["a", "b"]
    default_map = {"a": 1, "b": 2}
    default_record = {"sub_int": -3}
    schema = {
        "type": "record",
        "name": "test_all_default_values",
        "fields": [
            {"name": "boolean", "type": "boolean", "default": default_boolean},
            {"name": "string", "type": "string", "default": default_string},
            {"name": "bytes", "type": "bytes", "default": default_bytes},
            {"name": "int", "type": "int", "default": default_int},
            {"name": "long", "type": "long", "default": default_long},
            {"name": "float", "type": "float", "default": default_float},
            {"name": "double", "type": "double", "default": default_double},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
                "default": default_fixed,
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
                "default": default_union,
            },
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
                "default": default_enum,
            },
            {
                "name": "array",
                "type": {"type": "array", "items": "string"},
                "default": deepcopy(default_array),
            },
            {
                "name": "map",
                "type": {"type": "map", "values": "int"},
                "default": deepcopy(default_map),
            },
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
                "default": default_record,
            },
        ],
    }

    record = {}

    new_file = StringIO(json.dumps(record))
    read_record = next(json_reader(new_file, schema))

    assert read_record["boolean"] == default_boolean
    assert read_record["string"] == default_string
    assert read_record["bytes"] == default_bytes.encode("iso-8859-1")
    assert read_record["int"] == default_int
    assert read_record["long"] == default_long
    assert read_record["float"] == default_float
    assert read_record["double"] == default_double
    assert read_record["fixed"] == default_fixed.encode("iso-8859-1")
    assert read_record["union"] == default_union
    assert read_record["enum"] == default_enum
    assert read_record["array"] == default_array
    assert read_record["map"] == default_map
    assert read_record["record"] == default_record


def test_default_value_missing():
    """https://github.com/fastavro/fastavro/issues/485"""
    schema = {
        "type": "record",
        "name": "test_default_value_missing",
        "fields": [{"name": "string", "type": "string"}],
    }

    record = {}

    new_file = StringIO(json.dumps(record))
    with pytest.raises(ValueError, match="no value and no default"):
        next(json_reader(new_file, schema))


def test_map_of_union_of_array_and_map():
    """https://github.com/fastavro/fastavro/issues/572"""
    schema = {
        "name": "Test",
        "namespace": "test",
        "type": "record",
        "fields": [
            {
                "name": "metadata",
                "type": {
                    "type": "map",
                    "values": [
                        {"type": "array", "items": "string"},
                        {"type": "map", "values": ["string"]},
                    ],
                },
            }
        ],
    }

    records = [{"metadata": {"map1": {"map2": "str"}}}]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_json_writer_with_validation():
    """https://github.com/fastavro/fastavro/issues/580"""
    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
        ],
    }

    records = [
        {"station": "011990-99999", "temp": 0, "time": 1433269388},
        {"station": "011990-99999", "temp": 22, "time": "last day"},
        {"station": "011990-99999", "temp": -11, "time": 1433273379},
        {"station": "012650-99999", "temp": 111.9, "time": 1433275478},
    ]

    new_file = StringIO()
    with pytest.raises(ValidationError):
        json_writer(new_file, schema, records, validator=True)


def test_custom_encoder_and_decoder():
    """https://github.com/fastavro/fastavro/pull/579"""

    class CustomJSONEncoder(AvroJSONEncoder):
        """Encoder that will prepend an underscore to all string values"""

        def write_utf8(self, value):
            self._parser.advance(String())
            if self._parser.stack[-1] == MapKeyMarker():
                self._parser.advance(MapKeyMarker())
                self.write_object_key(value)
            else:
                self.write_value("_" + value)

    class CustomJSONDecoder(AvroJSONDecoder):
        """Decoder that will prepend an underscore to all string values"""

        def read_utf8(self):
            symbol = self._parser.advance(String())
            if self._parser.stack[-1] == MapKeyMarker():
                self._parser.advance(MapKeyMarker())
                for key in self._current:
                    self._key = key
                    break
                return self._key
            else:
                return "_" + self.read_value(symbol)

    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
        ],
    }

    records = [
        {"station": "011990-99999", "temp": 0, "time": 1433269388},
        {"station": "011990-99999", "temp": 22, "time": 1433269389},
        {"station": "011990-99999", "temp": -11, "time": 1433273379},
        {"station": "012650-99999", "temp": 111.9, "time": 1433275478},
    ]

    new_file = StringIO()
    json_writer(new_file, schema, records, encoder=CustomJSONEncoder)
    assert "_011990-99999" in new_file.getvalue()

    new_file.seek(0)
    for record in json_reader(new_file, schema, decoder=CustomJSONDecoder):
        assert record["station"].startswith("__")


def test_strict_option():
    """https://github.com/fastavro/fastavro/issues/549"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [
            {"name": "field_1", "type": "boolean"},
            {"name": "field_2", "type": ["null", "string"], "default": None},
        ],
    }

    test_record1 = {"field_1": True, "field_2": "foo", "field_3": "something"}
    test_record2 = {"field_1": True}
    test_record3 = {"field_2": "foo"}

    with pytest.raises(ValueError, match="field_3"):
        roundtrip(schema, [test_record1], writer_kwargs={"strict": True})

    with pytest.raises(ValueError, match="field_2 is specified .*? but missing"):
        roundtrip(schema, [test_record2], writer_kwargs={"strict": True})

    with pytest.raises(ValueError, match="field_1 is specified .*? but missing"):
        roundtrip(schema, [test_record3], writer_kwargs={"strict": True})


def test_strict_allow_default_option():
    """https://github.com/fastavro/fastavro/issues/549"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [
            {"name": "field_1", "type": "boolean"},
            {"name": "field_2", "type": ["null", "string"], "default": None},
        ],
    }

    test_record1 = {"field_1": True, "field_2": "foo", "field_3": "something"}
    test_record2 = {"field_1": True}
    test_record3 = {"field_2": "foo"}

    with pytest.raises(ValueError, match="field_3"):
        roundtrip(schema, [test_record1], writer_kwargs={"strict_allow_default": True})

    roundtrip(schema, [test_record2], writer_kwargs={"strict_allow_default": True})

    with pytest.raises(ValueError, match="field_1 is specified .*? but missing"):
        roundtrip(schema, [test_record3], writer_kwargs={"strict_allow_default": True})


def test_json_with_map():
    """https://github.com/fastavro/fastavro/issues/629"""
    schema = {
        "type": "record",
        "name": "User",
        "fields": [
            {"name": "name", "type": "string"},
            {"name": "age", "type": "long"},
            {
                "name": "addresses",
                "type": {
                    "type": "map",
                    "values": {
                        "type": "record",
                        "name": "Address",
                        "fields": [
                            {"name": "street", "type": "string"},
                            {"name": "street_number", "type": "long"},
                        ],
                        "doc": "An Address",
                    },
                    "name": "address",
                },
            },
        ],
        "doc": "User with multiple Address",
    }

    payload = {
        "name": "TogYzVenzFrgcVunpkfo",
        "age": 5694,
        "addresses": {
            "HGbZkCCabEbwaTXjbTEA": {
                "street": "tNXiLgAYswaCPLazSfus",
                "street_number": 1316,
            }
        },
    }

    roundtrip(schema, [payload])


def test_disable_tuple_notation_option():
    """https://github.com/fastavro/fastavro/issues/548"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [
            {"name": "foo", "type": ["string", {"type": "array", "items": "string"}]}
        ],
    }

    new_records = roundtrip(
        schema,
        [{"foo": ("string", "0")}],
        writer_kwargs={"disable_tuple_notation": True},
    )
    assert new_records == [{"foo": ["string", "0"]}]


def test_strict_allow_default_bug():
    """https://github.com/fastavro/fastavro/issues/638"""
    schema = {
        "namespace": "namespace",
        "name": "name",
        "type": "record",
        "fields": [{"name": "some_field", "type": "string", "default": "test"}],
    }

    test_record = {"eggs": "eggs"}

    with pytest.raises(ValueError, match="record contains more fields .*? eggs"):
        roundtrip(schema, [test_record], writer_kwargs={"strict_allow_default": True})



import ast
import os
import re
import sys
from setuptools import setup, Extension

try:
    import Cython
except ImportError:
    ext = ".c"
else:
    ext = ".pyx"

ext_modules = []
if not hasattr(sys, "pypy_version_info"):
    ext_modules += [
        Extension("fastavro._read", ["fastavro/_read" + ext]),
        Extension("fastavro._schema", ["fastavro/_schema" + ext]),
        Extension("fastavro._write", ["fastavro/_write" + ext]),
        Extension("fastavro._validation", ["fastavro/_validation" + ext]),
        Extension("fastavro._logical_readers", ["fastavro/_logical_readers" + ext]),
        Extension("fastavro._logical_writers", ["fastavro/_logical_writers" + ext]),
    ]


def version():
    pyfile = "fastavro/__init__.py"
    with open(pyfile) as fp:
        data = fp.read()

    match = re.search("__version_info__ = (\(.*\))", data)
    assert match, f"cannot find version in {pyfile}"
    vinfo = ast.literal_eval(match.group(1))
    return ".".join(str(v) for v in vinfo)


setup(
    name="fastavro",
    version=version(),
    description="Fast read/write of AVRO files",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Miki Tebeka",
    author_email="miki.tebeka@gmail.com",
    license="MIT",
    url="https://github.com/fastavro/fastavro",
    packages=["fastavro", "fastavro.io", "fastavro.repository"],
    ext_modules=ext_modules,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "fastavro = fastavro.__main__:main",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    python_requires=">=3.7",
    extras_require={
        "codecs": ["python-snappy", "zstandard", "lz4"],
        "snappy": ["python-snappy"],
        "zstandard": ["zstandard"],
        "lz4": ["lz4"],
    },
    package_data={"fastavro": ["py.typed"]},
)
